#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "nlllib.h"
#include "line.h"
#include "readline.h"
#include "nll.h"
#include "nllmain.h"

static void truncate_line(char *buffer)
{
  int len;
  if (buffer) {
    len = strlen(buffer);
    while (len > 0) {
      if (!isspace(buffer[len - 1]))
	break;
      len--;
    }
    buffer[len] = '\0';
  }
}

int nll_main(int quit, int auto_run, int auto_number, char *command)
{
  char *linebuf, *readlinebuf, *top, *p, *head;
  int finished = 0, r, number, number2, retcode = 0;
  line_t line;
  FILE *fp = NULL;
  char buffer[LINE_MAXLEN + 1 + 1]; /* +1 is for auto number space */

  while (!finished) {
    fp = nll_file_fp();
    if (!fp)
      auto_number = 0;

    linebuf = NULL;
    readlinebuf = NULL;

    if (fp) {
      linebuf = buffer + (auto_number ? 1 : 0);
      if (!fgets(linebuf, LINE_MAXLEN + 1, fp)) {
	nll_file_close();
	continue;
      }
      truncate_line(linebuf);
    } else if (command) {
      linebuf = command;
      command = NULL;
    } else if (auto_run) {
      linebuf = "RUN";
      auto_run = 0;
    } else if (quit) {
      finished = 1;
      continue;
    } else if (nll_is_tty()) {
      readlinebuf = nll_readline("nll> ");
      if (!readlinebuf) {
	fprintf(stdout, "\n");
	continue;
      }
      linebuf = readlinebuf;
      truncate_line(linebuf);
    } else {
      finished = 1;
      continue;
    }

    p = linebuf;
    while (isspace(*p))
      p++;

    head = p;
    number = strtoul(p, &top, 10);

    if (!auto_number && (top == p)) { /* direct mode */
      if ((r = nll_line_edit(-1, -1, top, &line)) < 0)
	NLL_ERREXIT(r, top);
      if (line) {
	nll_initialize_clear();
	nll_edit_clear();
	r = nll_parse(line);
	if (readlinebuf)
	  nll_add_history(nll_line_buffer(line));
	if (r >= 0) {
	  if ((r = nll_start(line, NULL, &retcode)) < 0)
	    finished = 1;
	}
	if ((r = nll_line_free(line)) < 0)
	  NLL_ERREXIT(r, top);
	if (nll_is_initialize()) {
	  if ((r = nll_new()) < 0)
	    NLL_ERREXIT(r, top);
	}
	if ((number = nll_edit()) >= 0) {
	  if ((r = line_get(number, &line)) < 0)
	    return r;
	  if (!line)
	    line = line_tail();
	  if (line)
	    nll_edit_line(number, line);
	}
      }
    } else { /* program mode */
      if (top == p) {
	number = 0;
	*(--top) = *p ? ' ' : '#';
	number2 = -1;
      } else {
	p = top;
	while (isspace(*p))
	  p++;
	number2 = isdigit(*p) ? strtoul(p, &top, 10) : -1;
      }
      if ((r = nll_line_edit(number, number2, top, &line)) < 0)
	NLL_ERREXIT(r, top);
      if (line)
	nll_parse(line);
      if (readlinebuf) {
	if (line)
	  strcpy(top, nll_line_buffer(line));
	nll_add_history(head);
      }
    }

    if (readlinebuf)
      free(readlinebuf);
  }

  return retcode;
}
